/*
 * @(#)JNIAltivec.c created 12/20/2002 Berkeley, Office 405
 *
 * Copyright (c) 2002 Luca Lutterotti All Rights Reserved.
 *
 * This software is the research result of Luca Lutterotti and it is 
 * provided as it is as confidential and proprietary information.  
 * You shall not disclose such Confidential Information and shall use 
 * it only in accordance with the terms of the license agreement you 
 * entered into with the author.
 *
 * THE AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */

#include "it_unitn_maud_JNIAltivec.h"
#if defined(__APPLE__) && defined(__MACH__)
/* Mac OS X (client) 10.x (or server 1.x/10.x) - gcc or Metrowerks MachO compilers */
	#define MAC_OS_X
#endif
#ifdef MAC_OS_X
	#include <sys/sysctl.h>
	#include <vecLib/vecLib.h>
#endif
#include <Math.h>

#define __full_copy

#ifdef MAC_OS_X
	typedef union {
		unsigned char c[16];
		signed char sc[16];
		unsigned short s[8];
		signed short ss[8];
		unsigned long l[4];
		signed long sl[4];
		float	f[4];
		vector unsigned char vc;
		vector unsigned short vs;
		vector unsigned int vi;
		vector float vf;
	} valign;
#endif

double sqrtln2pi = 4.6971863935E-01;
double LN2 = -6.9314718056E-01;
double PI = 3.14159265359;
double DEGTOPI = 0.01745329251994444;
double PITODEG = 57.29577951307855;
#ifdef MAC_OS_X

	vector float vLN2 = (vector float) (-6.9314718056E-01);	
	vector float vZero = (vector float) (0);
	vector float vOne = (vector float) (1);
	vector float vDEGTOPI = (vector float) (0.01745329251994444);	
	vector float vPITODEG = (vector float) (57.29577951307855);	

Boolean IsAltiVecAvailable( void ) {

	int selectors[2] = { CTL_HW, HW_VECTORUNIT };
	int hasVectorUnit = 0;
	size_t length = sizeof(hasVectorUnit);
	int error = sysctl(selectors, 2, &hasVectorUnit, &length, NULL, 0); 

	if( 0 == error ) return hasVectorUnit != 0;

	return FALSE;

}

/*********************************************************************************
*	vSelElemFP uses a union type to insure that the floating point array is  *
*	aligned on a 16-byte boundary.					         *
*********************************************************************************/
float vSelElemFP ( vector float x, long int n ) {
	valign xAlign;

/*********************************************************************************
*	Since xAlign is 16-byte aligned, we can store x into it and then         *
*	directly return the nth element.				         *
*********************************************************************************/
	xAlign.vf = x;
	
	return xAlign.f[n];
}

/*********************************************************************************
*	vInsElemFP uses a union type to insure that the floating point array is  *
*	aligned on a 16-byte boundary.					         *
*********************************************************************************/

vector float vInsElemFP ( vector float x, float xElem, long int n ) {
	valign xAlign;

	xAlign.vf = x;
	
/**********************************************************************************
*	Because of the union structure, we can store xElem directly into the nth  *
*	element of the vector x.						  *
**********************************************************************************/
	xAlign.f[n] = xElem;
	
	return xAlign.vf;
}

vector float vInsFP ( vector float x, float *xElem ) {
	valign xAlign;
	int n;
	
	xAlign.vf = x;

	for (n = 0; n < 4; n++)
		xAlign.f[n] = *(xElem + n);
	
	return xAlign.vf;
}
	
float *fpInsV ( vector float x, float *xElem ) {
	valign xAlign;
	int n;

	xAlign.vf = x;
	
	for (n = 0; n < 4; n++)
		*(xElem + n) = xAlign.f[n];
	
	return xElem;
}

vector float vCreateInsFP ( float *xElem ) {
	valign xAlign;
	int n;
	vector float x;
	
	x = (vector float) (0.0, 0.0, 0.0, 0.0);
	
	xAlign.vf = x;

	for (n = 0; n < 4; n++)
		xAlign.f[n] = *(xElem + n);
	
	return xAlign.vf;
}
	
float *fpCreateInsV ( vector float x ) {
	valign xAlign;
	int n;
	float *resultFP;

	xAlign.vf = x;
	
	resultFP = (float *) calloc(4, sizeof (float));
	for (n = 0; n < 4; n++)
		*(resultFP + n) = xAlign.f[n];
	
	return resultFP;
}

#endif

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    IsAltiVecAvailable
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_it_unitn_maud_JNIAltivec_IsAltiVecAvailable
  (JNIEnv *env, jclass obj) {
	jboolean isAvailable = JNI_FALSE;

#ifdef MAC_OS_X
	Boolean isAltivecAvailable = IsAltiVecAvailable();
	if (isAltivecAvailable == TRUE)
		isAvailable = JNI_TRUE;	
#endif

	return isAvailable;
}

#ifdef __cplusplus
  #define JNIGetArrayLength(env, jx) (env)->GetArrayLength(jx)
  #define JNIArrayCriticalGet(env, jx) (env)->GetPrimitiveArrayCritical(jx, 0)
  #define JNIArrayCriticalPut(env, jx, x) (env)->ReleasePrimitiveArrayCritical(jx, x, JNI_COMMIT)
  #define JNIArrayCriticalRel(env, jx, x) (env)->ReleasePrimitiveArrayCritical(jx, x, JNI_ABORT)
	#define JNINewFloatArray(env, size) (env)->NewFloatArray(size)
	#define JNINewIntArray(env, size) (env)->NewIntArray(size)
	#define JNISetFloatArrayRegion(env, fvC, size, vC) (env)->SetFloatArrayRegion(fvC, 0, size, vC)
	#define JNISetIntArrayRegion(env, fvC, size, vC) (env)->SetIntArrayRegion(fvC, 0, size, vC)
	#define JNIGetFloatArrayElements(env, fvA) (env)->GetFloatArrayElements(fvA, 0)
	#define JNIGetIntArrayElements(env, fvA) (env)->GetIntArrayElements(fvA, 0)
  #define JNIReleaseFloatArrayElements(env, fvB, vB) (env)->ReleaseFloatArrayElements(fvB, vB, JNI_ABORT);
  #define JNIReleaseIntArrayElements(env, fvB, vB) (env)->ReleaseIntArrayElements(fvB, vB, JNI_ABORT);
#else /* __cplusplus */
  #define JNIGetArrayLength(env, x) (*env)->GetArrayLength(env, x)
  #define JNIArrayCriticalGet(env, x) (*env)->GetPrimitiveArrayCritical(env, x, 0)
  #define JNIArrayCriticalPut(env, jx, x) (*env)->ReleasePrimitiveArrayCritical(env, jx, x, JNI_COMMIT)
  #define JNIArrayCriticalRel(env, jx, x) (*env)->ReleasePrimitiveArrayCritical(env, jx, x, JNI_ABORT)
	#define JNINewFloatArray(env, size) (*env)->NewFloatArray(env, size)
	#define JNINewIntArray(env, size) (*env)->NewIntArray(env, size)
	#define JNISetFloatArrayRegion(env, fvC, size, vC) (*env)->SetFloatArrayRegion(env, fvC, 0, size, vC)
	#define JNISetIntArrayRegion(env, fvC, size, vC) (*env)->SetIntArrayRegion(env, fvC, 0, size, vC)
	#define JNIGetFloatArrayElements(env, fvA) (*env)->GetFloatArrayElements(env, fvA, 0)
	#define JNIGetIntArrayElements(env, fvA) (*env)->GetIntArrayElements(env, fvA, 0)
  #define JNIReleaseFloatArrayElements(env, fvB, vB) (*env)->ReleaseFloatArrayElements(env, fvB, vB, JNI_ABORT);
  #define JNIReleaseIntArrayElements(env, fvB, vB) (*env)->ReleaseIntArrayElements(env, fvB, vB, JNI_ABORT);
#endif /* __cplusplus */

#define min(a, b) (a > b) ? b : a

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vdivf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vdivf
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vdivf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = fA[j] / fB[j];
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sdivf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sdivf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = vA[j] / vB[j];
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vsqrtf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vsqrtf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vsqrtf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = sqrt(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    ssqrtf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_ssqrtf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = sqrt(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vrsqrtf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vrsqrtf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vrsqrtf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = 1.0f / sqrt(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    srsqrtf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_srsqrtf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = 1.0f / sqrt(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vexpf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vexpf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vexpf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = exp(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sexpf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sexpf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = exp(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vexpm1f
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vexpm1f
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vexpm1f(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = exp(fA[j]) - 1.0f;
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sexpm1f
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sexpm1f
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = exp(vA[j]) - 1.0f;
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vlogf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vlogf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vlogf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = log(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    slogf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_slogf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = log(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vlog1pf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vlog1pf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vlog1pf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = log1p(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    slog1pf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_slog1pf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = log1p(vA[j]);  // log(1.0 + vX[j])
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vlogbf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vlogbf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vlogbf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = logb(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    slogbf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_slogbf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = logb(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vscalbf
 * Signature: ([F[I[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vscalbf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, ibsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		ibsize = size * sizeof(int);
		int remaining = osize - size;
		
		vector float *vA, *vC, *vT, *vM;
		vector signed int *vB, *vTi, *vMi;
		float *fA, *fC;
		int *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);
		fB = (int *)JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		vMi = (vector signed int *) malloc(ibsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vMi, fA, ibsize);
		vC = vA + bsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vscalbf(vA[i], vB[i]);
		}

		memcpy(fB, vB, ibsize);
		if (remaining > 0)
			for (i = 0; i < remaining; i++)
				fC[i + bsize] = scalb(fA[i + bsize], fB[i + ibsize]);

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
		free(vMi);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sscalbf
 * Signature: ([F[I[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sscalbf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB, jfloatArray fvC) {

	float *vA, *vC;
	int *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	vB = (int *) JNIArrayCriticalGet(env, fvB);
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *) JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = scalb(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vfabf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vfabf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vfabf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
// not availalble				fB[j] = fab(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sfabf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sfabf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
// not availalble			vB[j] = fab(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vcopysignf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vcopysignf
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vcopysignf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = copysign(fA[j], fB[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    scopysignf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_scopysignf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = copysign(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vnextafterf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vnextafterf
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vnextafterf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = nextafter(fA[j], fB[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    snextafterf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_snextafterf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = nextafter(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vclassifyf
 * Signature: ([F[I)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vclassifyf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, ibsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		ibsize = size * sizeof(int);
		int remaining = osize - size;
		
		vector float *vA, *vT, *vM;
		vector unsigned int *vB, *vMi;
		float *fA;
		int *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (int *) malloc(osize * sizeof(int));
		else
      fB = (int *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize);
		vMi = (vector unsigned int *) malloc(ibsize);
		memcpy(vA = vM, fA, bsize);
		vB = vMi;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vclassifyf(vA[i]);
		}

		memcpy(fB, vB, ibsize);
		if (remaining > 0)
			for (i = 0; i < remaining; i++)
// not availalble				fB[i + ibsize] = classify(fA[i + bsize]);

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewIntArray(env, osize);
			JNISetIntArrayRegion(env, fvB, osize, (jint *) fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
		free(vMi);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sclassifyf
 * Signature: ([F[I)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sclassifyf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB) {

	float *vA;
	int *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (int *) malloc(size * sizeof(int));
		else
      vB = (int *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
// not availalble			vB[j] = classify(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewIntArray(env, size);
			JNISetIntArrayRegion(env, fvB, size, (jint *) vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vsignbitf
 * Signature: ([F[I)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vsignbitf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, ibsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		ibsize = size * sizeof(int);
		int remaining = osize - size;
		
		vector float *vA, *vT, *vM;
		vector unsigned int *vB, *vTi, *vMi;
		float *fA;
		int *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (int *) malloc(osize * sizeof(int));
		else
      fB = (int *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize);
		vMi = (vector unsigned int *) malloc(ibsize);
		memcpy(vA = vM, fA, bsize);
		vB = vMi;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vsignbitf(vA[i]);
		}

		memcpy(fB, vB, ibsize);
		if (remaining > 0)
			for (i = 0; i < remaining; i++)
				fB[i + ibsize] = signbit(fA[i + bsize]);

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewIntArray(env, osize);
			JNISetIntArrayRegion(env, fvB, osize, (jint *) fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
		free(vMi);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    ssignbitf
 * Signature: ([F[I)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_ssignbitf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB) {

	float *vA;
	int *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (int *) malloc(size * sizeof(int));
		else
      vB = (int *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = signbit(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewIntArray(env, size);
			JNISetIntArrayRegion(env, fvB, size, (jint *) vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vsinf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vsinf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vsinf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = sin(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    ssinf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_ssinf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = sin(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vcosf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vcosf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vcosf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = cos(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    scosf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_scosf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = cos(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vtanf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vtanf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vtanf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = tan(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    stanf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_stanf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = tan(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vasinf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vasinf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vasinf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = asin(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sasinf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sasinf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = asin(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vacosf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vacosf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vacosf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = acos(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sacosf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sacosf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = acos(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vatanf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vatanf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vatanf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = atan(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    satanf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_satanf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = atan(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vatan2f
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vatan2f
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vatan2f(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = atan2(fA[j], fB[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    satan2f
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_satan2f
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = atan2(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vsinhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vsinhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vsinhf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = sinh(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    ssinhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_ssinhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = sinh(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vcoshf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vcoshf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vcoshf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = cosh(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    scoshf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_scoshf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = cosh(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vtanhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vtanhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vtanhf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = tanh(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    stanhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_stanhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = tanh(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vasinhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vasinhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vasinhf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = asinh(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sasinhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sasinhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = asinh(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vacoshf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vacoshf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vacoshf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = acosh(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sacoshf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sacoshf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = acosh(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vatanhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vatanhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = JNIGetArrayLength(env, fvA);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vT;
		vector float *vM;
		float *fA, *fB;

		fA = (float *)JNIArrayCriticalGet(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < osize)
			fB = (float *) malloc(osize * sizeof(float));
		else
      fB = (float *)JNIArrayCriticalGet(env, fvB);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *) malloc(bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		vB = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vB[i] = vatanhf(vA[i]);
		}

		memcpy(fB, vB, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fB[j] = atanh(fA[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		if (fvB == NULL || csize < osize) {
			fvB = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvB, osize, fB);
			free(fB);
		} else
			JNIArrayCriticalPut(env, fvB, fB);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    satanhf
 * Signature: ([F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_satanhf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB) {

	float *vA, *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL) {
  	vA = (float *) JNIArrayCriticalGet(env, fvA);
  	size = JNIGetArrayLength(env, fvA);

		if (fvB == NULL || (csize = JNIGetArrayLength(env, fvB)) < size)
			vB = (float *) malloc(size * sizeof(float));
		else
      vB = (float *) JNIArrayCriticalGet(env, fvB);

		for (j = 0; j < size; j++) {
			vB[j] = atanh(vA[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		if (fvB == NULL || csize < size) {
			fvB = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvB, size, vB);
			free(vB);
		} else
			JNIArrayCriticalPut(env, fvB, vB);
	}

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vfmodf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vfmodf
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vfmodf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = fmod(fA[j], fB[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sfmodf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sfmodf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = fmod(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vremainderf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vremainderf
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vremainderf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = remainder(fA[j], fB[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sremainderf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sremainderf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = remainder(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vremquof
 * Signature: ([F[F[I[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vremquof
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jintArray ivn, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, ibsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL && ivn != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		jsize shouldBe4u = sizeof(vector unsigned int) / sizeof(unsigned int);
		// in case the lengths are differents
    osize = min(min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB)), JNIGetArrayLength(env, ivn));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
		jsize usize = vsize * shouldBe4u;  
    bsize = size * sizeof(float);
    ibsize = usize * sizeof(unsigned int);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC;
		vector unsigned int *vN, *vMi;
		vector float *vM;
		float *fA, *fB, *fC;
		unsigned int *iN;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);
		iN = (unsigned int *) JNIArrayCriticalGet(env, ivn);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		vMi = (vector unsigned int *) malloc(ibsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		memcpy(vN = vMi, iN, ibsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vremquof(vA[i], vB[i], &vN[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = remquo(fA[j], fB[j], &iN[i + ibsize]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		JNIArrayCriticalPut(env, ivn, iN);
		JNIArrayCriticalRel(env, ivn, iN);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
		free(vMi);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sremquof
 * Signature: ([F[F[I[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sremquof
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray ivN, jfloatArray fvC) {

  float *vA, *vB, *vC;
	unsigned int *vN;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB)), JNIGetArrayLength(env, ivN));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
  	vN = (unsigned int *)JNIArrayCriticalGet(env, ivN);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = remquo(vA[j], vB[j], &vN[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		JNIArrayCriticalPut(env, ivN, vN);
		JNIArrayCriticalRel(env, ivN, vN);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vipowf
 * Signature: ([F[I[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vipowf
(JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, ibsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
    ibsize = size * sizeof(int);
		int remaining = osize - size;
		
		vector float *vA, *vC, *vT;
		vector signed int *vB;
		vector float *vM;
		vector signed int *vMi;
		float *fA, *fC;
		int *fB;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (int *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize);
		vMi = (vector signed int *)malloc(ibsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vMi, fB, ibsize);
		vC = vA + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vipowf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = pow(fA[j], fB[i + ibsize]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
		free(vMi);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sipowf
 * Signature: ([F[I[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_sipowf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jintArray fvB, jfloatArray fvC) {

  float *vA, *vC;
	int *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (int *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = pow(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vpowf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vpowf
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize;
  int i;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		vector float *vA, *vB, *vC, *vT;
		vector float *vM;
		float *fA, *fB, *fC;

		fA = (float *) JNIArrayCriticalGet(env, fvA);
		fB = (float *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (float *) malloc(osize * sizeof(float));
		else
      fC = (float *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(bsize + bsize + bsize);
		memcpy(vA = vM, fA, bsize);
		memcpy(vB = vA + vsize, fB, bsize);
		vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0; i < vsize; i++) {
			vC[i] = vpowf(vA[i], vB[i]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;
				fC[j] = pow(fA[j], fB[j]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewFloatArray(env, osize);
			JNISetFloatArrayRegion(env, fvC, osize, fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    spowf
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_spowf
  (JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {

  float *vA, *vB, *vC;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (float *)JNIArrayCriticalGet(env, fvA);
  	vB = (float *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (float *) malloc(size * sizeof(float));
		else
      vC = (float *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
			vC[j] = pow(vA[j], vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewFloatArray(env, size);
			JNISetFloatArrayRegion(env, fvC, size, vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vtablelookup
 * Signature: ([I[J[I)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vtablelookup
(JNIEnv *env, jclass obj, jintArray fvA, jlongArray fvB, jintArray fvC) {
#ifdef MAC_OS_X
  jsize osize, size, bsize, vsize, csize, lvsize, lsize, lbsize;
  int i, j;

  if (fvA != NULL && fvB != NULL) {
		jsize shouldBe4 = sizeof(vector signed int) / sizeof(signed int);
		// in case the lengths are differents
    osize = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(signed int);
		int remaining = osize - size;
    lbsize = size * sizeof(long);
		
		vector signed int *vA;
		vector unsigned int *vC;
		vector signed int *vM;
		vector unsigned int *vMu;
		signed int *fA;
		unsigned int *fC;
		long *fB;

		fA = (signed int *) JNIArrayCriticalGet(env, fvA);
		fB = (long *) JNIArrayCriticalGet(env, fvB);

		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < osize)
			fC = (unsigned int *) malloc(osize * sizeof(unsigned int));
		else
      fC = (unsigned int *)JNIArrayCriticalGet(env, fvC);
			
        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector signed int *)malloc(bsize);
		vMu = (vector unsigned int *)malloc(bsize);
		memcpy(vA = vM, fA, bsize);
		vC = vMu;

        // Could unroll this loop since we likely have two vector units.
		for (i = 0, j = 0; i < vsize; i++, j += shouldBe4) {
			vC[i] = vtablelookup(vA[i], &fB[j]);
		}

		memcpy(fC, vC, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				j = i + bsize;
// not availalble				fC[j] = tablelookup(fA[j], &fB[i + lbsize]);
			}
		}

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalRel(env, fvA, fA);
		JNIArrayCriticalRel(env, fvB, fB);
		if (fvC == NULL || csize < osize) {
			fvC = JNINewIntArray(env, osize);
			JNISetIntArrayRegion(env, fvC, osize, (jint *) fC);
			free(fC);
		} else
			JNIArrayCriticalPut(env, fvC, fC);

		free(vM);
		free(vMu);
  }
#endif
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    stablelookup
 * Signature: ([I[J[I)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_stablelookup
  (JNIEnv *env, jclass obj, jintArray fvA, jlongArray fvB, jintArray fvC) {

  signed int *vA;
	unsigned int *vC;
	long *vB;
  jsize size, csize;
  int i, j;
  
	if(fvA != NULL && fvB != NULL) {
  	size = min(JNIGetArrayLength(env, fvA), JNIGetArrayLength(env, fvB));
  	vA = (signed int *)JNIArrayCriticalGet(env, fvA);
  	vB = (long *)JNIArrayCriticalGet(env, fvB);
		if (fvC == NULL || (csize = JNIGetArrayLength(env, fvC)) < size)
			vC = (unsigned int *) malloc(size * sizeof(unsigned int));
		else
      vC = (unsigned int *)JNIArrayCriticalGet(env, fvC);

		for (j = 0; j < size; j++) {
// not availalble			vC[j] = tablelookup(vA[j], &vB[j]);
		}

		JNIArrayCriticalRel(env, fvA, vA);
		JNIArrayCriticalRel(env, fvB, vB);
		if (fvC == NULL || csize < size) {
			fvC = JNINewIntArray(env, size);
			JNISetIntArrayRegion(env, fvC, size, (jint *) vC);
			free(vC);
		} else
			JNIArrayCriticalPut(env, fvC, vC);
	}
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vpseudovoigt
 * Signature: ([FIIFFFF)[F
 */
JNIEXPORT jfloatArray JNICALL Java_it_unitn_maud_JNIAltivec_vpseudovoigt
  (JNIEnv *env, jclass obj, jfloatArray x, jint minindex, jint maxindex,
   jfloat intensity, jfloat eta, jfloat hwhm, jfloat position) {
#ifdef MAC_OS_X
	union {
		float scalars[vec_step(vector float)];
		vector float v;
	} alignedVFX, alignedVFC, vDGX, vDCX, vPOS, vHWHM;
	
  jfloat *vX, *vC;
  jfloat *vXp, *vCp;
  jfloatArray fvC;
  jsize size, size4;
  int i;
  float dgx, dcx;
  
	if(x != NULL) {
		size = maxindex - minindex;
#ifdef __full_copy
  #ifdef __cplusplus
  	vX = env->GetFloatArrayElements(x, 0);
  #else
  	vX = (*env)->GetFloatArrayElements(env, x, 0);
  #endif
#else
  #ifdef __cplusplus
  	env->GetFloatArrayRegion(x, minindex, size, vX);
  #else
  	(*env)->GetFloatArrayRegion(env, x, minindex, size, vX);
  #endif
#endif
		
		size4 = size / 4;
		vC = (jfloat *) calloc(size, sizeof(jfloat));
		vCp = vC;
#ifdef __full_copy
		vXp = vX+minindex;
#else
		vXp = vX;
#endif
		
		dgx = intensity * (1.0 - eta) * sqrtln2pi / hwhm;
		dcx = intensity * eta / (PI * hwhm);
		
		for (i = 0; i < 4; i++) {
			vDGX.scalars[i] = dgx;
			vDCX.scalars[i] = dcx;
			vPOS.scalars[i] = position;
			vHWHM.scalars[i] = hwhm;
		}
		vHWHM.v = vec_re(vHWHM.v);

		for (i = 0; i < size4; i++) {
			alignedVFX.scalars[0] = *(vXp++);
			alignedVFX.scalars[1] = *(vXp++);
			alignedVFX.scalars[2] = *(vXp++);
			alignedVFX.scalars[3] = *(vXp++);
			
			alignedVFX.v = vec_sub(vPOS.v, alignedVFX.v);
			alignedVFX.v = vec_madd(vHWHM.v, alignedVFX.v, vZero);
			alignedVFX.v = vec_madd(alignedVFX.v, alignedVFX.v, vZero);
			alignedVFC.v = vec_add(vOne, alignedVFX.v);
			alignedVFC.v = vdivf(vDCX.v, alignedVFC.v);
			alignedVFX.v = vec_madd( vLN2, alignedVFX.v, vZero );
			alignedVFX.v = vexpf( alignedVFX.v );
			alignedVFC.v = vec_madd( vDGX.v, alignedVFX.v, alignedVFC.v);
			
			*(vCp++) = alignedVFC.scalars[0];
			*(vCp++) = alignedVFC.scalars[1];
			*(vCp++) = alignedVFC.scalars[2];
			*(vCp++) = alignedVFC.scalars[3];
		}

#ifdef __cplusplus
  #ifdef __full_copy
		env->ReleaseFloatArrayElements(x, vX, 0);
  #endif
		fvC = env->NewFloatArray(size);
		env->SetFloatArrayRegion(fvC, 0, size, vC);
#else
  #ifdef __full_copy
		(*env)->ReleaseFloatArrayElements(env, x, vX, 0);
  #endif
		fvC = (*env)->NewFloatArray(env, size);
		(*env)->SetFloatArrayRegion(env, fvC, 0, size, vC);
#endif
		free(vC);
		return fvC;
	}
#endif
  return NULL;
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vpseudovoigts
 * Signature: ([F[F[I[I[F[F[F[F)I
 */
JNIEXPORT jint JNICALL Java_it_unitn_maud_JNIAltivec_vpseudovoigts
  (JNIEnv *env, jclass obj, jfloatArray x, jfloatArray f, jintArray minindex, jintArray maxindex,
   jfloatArray intensity, jfloatArray eta, jfloatArray hwhm, jfloatArray position) {
	
#ifdef MAC_OS_X
	union {
		float scalars[vec_step(vector float)];
		vector float v;
	} vDGX, vDCX, vPOS, vHWHM;
	
  jfloat *vintensity, *veta, *vhwhm, *vposition;
  jint *vminindex, *vmaxindex;
  jsize size, size4, bsize, vsize, asize;
	jsize numberOfPV, numberOfPoints;
  int i, ipv, remainder;
  float dgx, dcx;
  
	if(x != NULL && f != NULL) {
//    size = JNIGetArrayLength(env, x);

		vector float *vX, *vF;
		vector float vT1, vT2;
		
		// Determine if the arrays are conveniently aligned already.
		vX = (vector float *)JNIGetFloatArrayElements(env, x);
		vF = (vector float *)JNIGetFloatArrayElements(env, f);
		
  	vminindex = JNIGetIntArrayElements(env, minindex);
  	vmaxindex = JNIGetIntArrayElements(env, maxindex);
  	vintensity = JNIGetFloatArrayElements(env, intensity);
  	veta = JNIGetFloatArrayElements(env, eta);
  	vhwhm = JNIGetFloatArrayElements(env, hwhm);
  	vposition = JNIGetFloatArrayElements(env, position);
  	numberOfPV = JNIGetArrayLength(env, minindex);
  	numberOfPoints = JNIGetArrayLength(env, f);
		
		for (ipv = 0; ipv < numberOfPV; ipv++) {
			int imin = (int) (vminindex[ipv] / 4);
			int imax = ((int) ((vmaxindex[ipv] - 1) / 4)) + 1;
//			vminindex[ipv] = imin * 4;
//			vmaxindex[ipv] = imax * 4;
// should never happen
//			if (vmaxindex[ipv] > size)
//				vmaxindex[ipv] -= 4;

			dgx = vintensity[ipv] * (1.0 - veta[ipv]) * sqrtln2pi / vhwhm[ipv];
			dcx = vintensity[ipv] * veta[ipv] / (PI * vhwhm[ipv]);
		
			for (i = 0; i < 4; i++) {
				vDGX.scalars[i] = dgx;
				vDCX.scalars[i] = dcx;
				vPOS.scalars[i] = vposition[ipv];
				vHWHM.scalars[i] = vhwhm[ipv];
			}
			vHWHM.v = vec_re(vHWHM.v);

			for (i = imin; i < imax; i++) {
				vT1 = vec_sub(vPOS.v, vX[i]);
				vT1 = vec_madd(vHWHM.v, vT1, vZero);
				vT1 = vec_madd(vT1, vT1, vZero);
				vT2 = vec_add(vOne, vT1);
				vT2 = vdivf(vDCX.v, vT2);
				vT2 = vec_add(vF[i], vT2);
				vT1 = vec_madd( vLN2, vT1, vZero );
				vT1 = vexpf( vT1 );
				vF[i] = vec_madd( vDGX.v, vT1, vT2);
			}
		}

		JNISetFloatArrayRegion(env, f, numberOfPoints, (jfloat *) vF);
		JNIReleaseFloatArrayElements(env, x, (jfloat *) vX);
		JNIReleaseFloatArrayElements(env, f, (jfloat *) vF);
		JNIReleaseIntArrayElements(env, minindex, vminindex);
		JNIReleaseIntArrayElements(env, maxindex, vmaxindex);
		JNIReleaseFloatArrayElements(env, intensity, vintensity);
		JNIReleaseFloatArrayElements(env, eta, veta);
		JNIReleaseFloatArrayElements(env, hwhm, vhwhm);
		JNIReleaseFloatArrayElements(env, position, vposition);

		return 0;
	}
#endif
  return 1;
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    spseudovoigt
 * Signature: ([FIIFFFF)[F
 */
JNIEXPORT jfloatArray JNICALL Java_it_unitn_maud_JNIAltivec_spseudovoigt
  (JNIEnv *env, jclass obj, jfloatArray x, jint minindex, jint maxindex,
   jfloat intensity, jfloat eta, jfloat hwhm, jfloat position) {

  jfloat *vX, *vXp, *vC, *vCp;
  jfloat dgx, dcx, dx;
  jfloatArray fvC;
  jsize size;
  int i, j;
  
	if(x != NULL) {
		size = maxindex - minindex;
#ifdef __full_copy
  #ifdef __cplusplus
  	vX = env->GetFloatArrayElements(x, 0);
  #else
  	vX = (*env)->GetFloatArrayElements(env, x, 0);
  #endif
#else
  #ifdef __cplusplus
  	env->GetFloatArrayRegion(x, minindex, size, vX);
  #else
  	(*env)->GetFloatArrayRegion(env, x, minindex, size, vX);
  #endif
#endif
		
		vC = (jfloat *) calloc(size, sizeof(jfloat));
		vCp = vC;
#ifdef __full_copy
		vXp = vX+minindex;
#else
		vXp = vX;
#endif

    dgx = intensity * (1.0f - eta) * sqrtln2pi / hwhm;
    dcx = intensity * eta / (PI * hwhm);
  	for (i = 0; i < size; i++) {
    	dx = (position - *(vXp++)) / hwhm;
      dx *= dx;
      if (dx > 1.0E30)
        *(vCp++) = 0.0f;
      else if (dx > 30.0) {
        *(vCp++) = dcx / (1.0 + dx);
      } else {
        *(vCp++) = dcx / (1.0 + dx) + dgx * exp(LN2 * dx);
      }
  	}

#ifdef __cplusplus
  #ifdef __full_copy
		env->ReleaseFloatArrayElements(x, vX, 0);
  #endif
		fvC = env->NewFloatArray(size);
		env->SetFloatArrayRegion(fvC, 0, size, vC);
#else
  #ifdef __full_copy
		(*env)->ReleaseFloatArrayElements(env, x, vX, 0);
  #endif
		fvC = (*env)->NewFloatArray(env, size);
		(*env)->SetFloatArrayRegion(env, fvC, 0, size, vC);
#endif
		free(vC);
		return fvC;
	}
  return NULL;
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    spseudovoigts
 * Signature: ([F[F[I[I[F[F[F[F)I
 */
JNIEXPORT jint JNICALL Java_it_unitn_maud_JNIAltivec_spseudovoigts
  (JNIEnv *env, jclass obj, jfloatArray x, jfloatArray f, jintArray minindex, jintArray maxindex,
   jfloatArray intensity, jfloatArray eta, jfloatArray hwhm, jfloatArray position) {
	
  float *vintensity, *veta, *vhwhm, *vposition, *vX, *vF;
  jint *vminindex, *vmaxindex;
  jsize size, size4, bsize, vsize, asize;
	jsize numberOfPV, numberOfPoints;
  int i, ipv, remainder;
  float dgx, dcx;
  
	if(x != NULL && f != NULL) {
//    size = JNIGetArrayLength(env, x);

		// Determine if the arrays are conveniently aligned already.
		vX = (float *)JNIGetFloatArrayElements(env, x);
		vF = (float *)JNIGetFloatArrayElements(env, f);
		
  	vminindex = JNIGetIntArrayElements(env, minindex);
  	vmaxindex = JNIGetIntArrayElements(env, maxindex);
  	vintensity = (float *)JNIGetFloatArrayElements(env, intensity);
  	veta = (float *)JNIGetFloatArrayElements(env, eta);
  	vhwhm = (float *)JNIGetFloatArrayElements(env, hwhm);
  	vposition = (float *)JNIGetFloatArrayElements(env, position);
  	numberOfPV = JNIGetArrayLength(env, minindex);
  	numberOfPoints = JNIGetArrayLength(env, f);
		
		for (ipv = 0; ipv < numberOfPV; ipv++) {
			int imin = (int) vminindex[ipv];
			int imax = (int) vmaxindex[ipv];

			dgx = vintensity[ipv] * (1.0 - veta[ipv]) * sqrtln2pi / vhwhm[ipv];
			dcx = vintensity[ipv] * veta[ipv] / (PI * vhwhm[ipv]);
		
			for (i = imin; i < imax; i++) {
				float dx = (vposition[ipv] - vX[i]) / vhwhm[ipv];
				dx *= dx;
				if (dx > 1.0E30)
					vF[i] = 0.0f;
				else if (dx > 30.0) {
        	vF[i] = dcx / (1.0f + dx);
				} else {
					vF[i] = dcx / (1.0f + dx) + dgx * (float) exp(LN2 * dx);
				}
			}
		}

		JNISetFloatArrayRegion(env, f, numberOfPoints, (float *) vF);
		JNIReleaseFloatArrayElements(env, x, (float *) vX);
		JNIReleaseFloatArrayElements(env, f, (float *) vF);
		JNIReleaseIntArrayElements(env, minindex, vminindex);
		JNIReleaseIntArrayElements(env, maxindex, vmaxindex);
		JNIReleaseFloatArrayElements(env, intensity, (float *) vintensity);
		JNIReleaseFloatArrayElements(env, eta, (float *) veta);
		JNIReleaseFloatArrayElements(env, hwhm, (float *) vhwhm);
		JNIReleaseFloatArrayElements(env, position, (float *) vposition);

		return 0;
	}
  return 1;
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vComputeTextureAngles
 * Signature: ([F[F[F[F[FF)I
 */
																									
JNIEXPORT jint JNICALL Java_it_unitn_maud_JNIAltivec_vComputeTextureAngles
  (JNIEnv *env, jclass obj, jfloatArray omega, jfloatArray chi, jfloatArray phi, jfloatArray eta,
   jfloatArray sampleAngles, jfloat position) {
	
#ifdef MAC_OS_X
	union {
		float scalars[vec_step(vector float)];
		vector float v;
	} sinTheta, cosTheta, sinOmegaSample, cosOmegaSample, sinCsiSample, cosCsiSample, samplePhi, 
		M13, M23, cosPsi, tmpPhi;

  jsize osize, size, bsize, vsize, csize;
  int i, j;
	
  float *fomega, *fchi, *fphi, *feta, *fsampleAngles;
  vector float *vM, *vomega, *vchi, *vphi, *veta;
	
	if(omega != NULL) {
		jsize shouldBe4 = sizeof(vector float) / sizeof(float);
		osize = JNIGetArrayLength(env, omega);
		
		// we take only the part usable as vector, avoiding paranoid check
		vsize = osize / shouldBe4;
		size = vsize * shouldBe4;  
    bsize = size * sizeof(float);
		int remaining = osize - size;
		
		fomega = (float *)JNIArrayCriticalGet(env, omega);
		fchi = (float *)JNIArrayCriticalGet(env, chi);
		fphi = (float *)JNIArrayCriticalGet(env, phi);
		feta = (float *)JNIArrayCriticalGet(env, eta);
		fsampleAngles = (float *)JNIArrayCriticalGet(env, sampleAngles);

        // Doing a single malloc and free avoids unnecessary overhead.
		vM = (vector float *)malloc(4 * bsize);
		memcpy(vomega = vM, fomega, bsize);
		memcpy(vchi = vomega + vsize, fchi, bsize);
		memcpy(vphi = vchi + vsize, fphi, bsize);
		memcpy(veta = vphi + vsize, feta, bsize);
		
		position /= 2.0f;
		position *= DEGTOPI;
		for (i = 0; i < 3; i++)
			fsampleAngles[i] *= DEGTOPI;
		float fsinOmegaSample = (float) sin(fsampleAngles[0]);
		float fcosOmegaSample = (float) cos(fsampleAngles[0]);
		float fsinCsiSample = (float) sin(fsampleAngles[1]);
		float fcosCsiSample = (float) cos(fsampleAngles[1]);
		float fsinTheta = (float) sin(position);
		float fcosTheta = (float) cos(position);
		for (i = 0; i < 4; i++) {
			sinTheta.scalars[i] = fsinTheta;
			cosTheta.scalars[i] = fcosTheta;
			sinOmegaSample.scalars[i] = fsinOmegaSample;
			cosOmegaSample.scalars[i] = fcosOmegaSample;
			sinCsiSample.scalars[i] = fsinCsiSample;
			cosCsiSample.scalars[i] = fcosCsiSample;
			samplePhi.scalars[i] = fsampleAngles[2];
		}

		for (i = 0; i < vsize; i++) {
			vomega[i] = vec_madd(vomega[i], vDEGTOPI, vZero);
			vchi[i] = vec_madd(vchi[i], vDEGTOPI, vZero);
			vphi[i] = vec_madd(vphi[i], vDEGTOPI, samplePhi.v);
			veta[i] = vec_madd(veta[i], vDEGTOPI, vZero);
						
			vector float sinOmega = vsinf(vomega[i]);
			vector float cosOmega = vcosf(vomega[i]);
			vector float sinPhi = vsinf(vphi[i]);
			vector float cosPhi = vcosf(vphi[i]);
			vector float sinCsi = vsinf(vchi[i]);
			vector float cosCsi = vcosf(vchi[i]);
			vector float sinEta = vsinf(veta[i]);
			vector float cosEta = vcosf(veta[i]);
		
			vector float  c1 = vec_madd(cosOmegaSample.v, cosPhi, vZero);
			vector float  cs1 = vec_madd(cosOmegaSample.v, sinPhi, vZero);
			vector float 	s1 = vec_madd(sinOmegaSample.v, sinPhi, vZero);
			vector float 	s2 = vec_madd(sinOmegaSample.v, cosPhi, vZero);
			vector float  c2 = vec_madd(cosCsiSample.v, cosCsi, vZero);
			vector float  c3 = vec_madd(cosCsiSample.v, sinCsi, vZero);
			vector float  c4 = vec_madd(sinCsiSample.v, sinCsi, vZero);
			vector float  const1 = vec_madd(c1, sinCsiSample.v, c2);
			vector float 	nc3 = vec_sub(vZero, const1); // -const1
			vector float  const2 = vec_madd(cosOmegaSample.v, c2, vZero);
			const2 = vec_madd(nc3, sinCsi, const2);
			vector float  const3 = vec_sub(vZero, s2);
			const3 = vec_madd(cs1, sinCsiSample.v, const3);
			vector float  const4 = vec_madd(cosOmegaSample.v, c3, vZero);
			const4 = vec_madd(const1, cosCsi, const4);
			vector float  const5 = vec_sub(vZero, cs1);
			const5 = vec_madd(s2, sinCsiSample.v, const5);
			vector float  const6 = vec_sub(vZero, const5);
			const6 = vec_madd(const6, sinCsi, vZero);
			const6 = vec_madd(sinOmegaSample.v, c2, const6);
			vector float  const7 = vec_madd(s1, sinCsiSample.v, c1);
			vector float  const8 = vec_madd(sinOmegaSample.v, c3, vZero);
			const8 = vec_madd(const5, cosCsi, const8);
			vector float  const9 = vec_madd(sinCsiSample.v, cosCsi, vZero);
			const9 = vec_nmsub(c3, cosPhi, const9);
			vector float  c5 = vec_madd(cosCsiSample.v, sinPhi, vZero);
			vector float  const10 = vec_madd(c5, sinOmega, vZero);
			const10 = vec_madd(const9, cosOmega, const10);
			vector float  const11 = vec_sub(vZero, c4);
			const11 = vec_madd(c2, cosPhi, const11);
    
			c5 = vec_sub(vZero, c5);
			c5 = vec_madd(c5, cosOmega, vZero);
			c1 = vec_madd(const9, sinOmega, c5);
			c1 = vec_madd(c1, sinTheta.v, vZero);
			c2 = vec_madd(const11, sinEta, vZero);
			c2 = vec_madd(const10, cosEta, c2);
			M13.v = (c2, cosTheta.v, c1);

			c3 = vec_sub(vZero, const7);
			c3 = vec_madd(c3, cosOmega, vZero);
			c3 = vec_madd(const6, sinOmega, c3);
			c3 = vec_madd(c3, sinTheta.v, vZero);
			c4 = vec_madd(const8, sinEta, vZero);
			c1 = vec_madd(const7, sinOmega, vZero);
			c1 = vec_madd(const6, cosOmega, c1);
			c1 = vec_madd(c1, cosEta, c4);
			M23.v = vec_madd(c1, cosTheta.v, c3);
			
			c3 = vec_sub(vZero, const3);
			c3 = vec_madd(c3, cosOmega, vZero);
			c3 = vec_madd(const2, sinOmega, c3);
			c3 = vec_madd(c3, sinTheta.v, vZero);
			c4 = vec_madd(const4, sinEta, vZero);
			c1 = vec_madd(const3, sinOmega, vZero);
			c1 = vec_madd(const2, cosOmega, c1);
			c1 = vec_madd(c1, cosEta, c4);
			cosPsi.v = vec_madd(c1, cosTheta.v, c3);
		
			cosPsi.v = vacosf(cosPsi.v);
			vomega[i] = vec_madd(cosPsi.v, vPITODEG, vZero);
			cosPsi.v = vomega[i];
			tmpPhi.v = vchi[i];
			
			for (j = 0; j < 4; j++) {
		
				if (fabs(M23.scalars[j]) < 1.0E-9f)
					M23.scalars[j] = 0.0f;
				if (fabs(M13.scalars[j]) < 1.0E-9f) {
					if (M23.scalars[j] >= 0.0f)
						tmpPhi.scalars[j] = 90.0f;
					else
						tmpPhi.scalars[j] = -90.0f;
				} else if (M23.scalars[j] == 0.0f) {
					if (M13.scalars[j] > 0.0f)
						tmpPhi.scalars[j] = 180.0f;
					else
						tmpPhi.scalars[j] = 0.0f;
				} else {
					tmpPhi.scalars[j] = - (float) (atan(M23.scalars[j] / M13.scalars[j]) * PITODEG);
					if (M13.scalars[j] > 0.0f)
						tmpPhi.scalars[j] += 180.0f;
				}
		
				if (cosPsi.scalars[j] < 0.0f) {
					cosPsi.scalars[j] = -cosPsi.scalars[j];
					tmpPhi.scalars[j] += 180.0f;
				}
				if (cosPsi.scalars[j] > 90.0f) {
					cosPsi.scalars[j] = 180.0f - cosPsi.scalars[j];
					tmpPhi.scalars[j] += 180.0f;
				}
				while (tmpPhi.scalars[j] > 360.0f)
					tmpPhi.scalars[j] -= 360.0f;
				while (tmpPhi.scalars[j] < 0.0f)
					tmpPhi.scalars[j] += 360.0f;
			}
		}
		memcpy(fomega, vomega, bsize);
		memcpy(fchi, vchi, bsize);
		if (remaining > 0) {
			for (i = 0; i < remaining; i++) {
				int j = i + bsize;

				fomega[j] += fsampleAngles[2];
				float sinOmega = sin(fomega[j]);
				float cosOmega = cos(fomega[j]);
				float sinPhi = sin(fphi[j]);
				float cosPhi = cos(fphi[j]);
				float sinCsi = sin(fchi[j]);
				float cosCsi = cos(fchi[j]);
				float sinEta = sin(feta[j]);
				float cosEta = cos(feta[j]);
		
				float c1 = fcosOmegaSample * cosPhi;
				float cs1 = fcosOmegaSample * sinPhi;
				float s2 = fsinOmegaSample * cosPhi;
				float c2 = fcosCsiSample * cosCsi;
				float c3 = fcosCsiSample * sinCsi;
				float c4 = fsinCsiSample * sinCsi;
				float s1 = fcosCsiSample * sinPhi;
				float const1 = c1 * fsinCsiSample + c2;
				float const2 = -const1 * sinCsi + fcosOmegaSample * c2;
				float const3 = cs1 * fsinCsiSample - s2;
				float const4 = const1 * cosCsi + fcosOmegaSample * c3;
				float const5 = s2 * fsinCsiSample - cs1;
				float const6 = -const5 * sinCsi + fsinOmegaSample * c2;
				float const7 = fsinOmegaSample * sinPhi * fsinCsiSample + c1;
				float const8 = const5 * cosCsi + fsinOmegaSample * c3;
				float const9 = -c3 * cosPhi - fsinCsiSample * cosCsi;
				float const10 = s1 * sinOmega + const9 * cosOmega;
				float const11 = c2 * cosPhi - c4;

				float M13 = -(s1 * cosOmega - const9 * sinOmega) * fsinTheta + 
										(const11 * sinEta + const10 * cosEta) * fcosTheta; // res[0][2];
				float M23 = -(const7 * cosOmega - const6 * sinOmega) * fsinTheta + (const8 * sinEta + 
										(const7 * sinOmega + const6 * cosOmega) * cosEta) * fcosTheta; // res[1][2];
				float cosPsi = -(const3 * cosOmega - const2 * sinOmega) * fsinTheta + (const4 * sinEta + 
											(const3 * sinOmega + const2 * cosOmega) * cosEta) * fcosTheta; // res[2][2];
		
				fomega[j] = (float) (acos(cosPsi) * PITODEG);
		
				if (fabs(M23) < 1.0E-9)
					M23 = 0.0f;
				if (fabs(M13) < 1.0E-9) {
					if (M23 >= 0.0)
						fchi[j] = 90.0f;
					else
						fchi[j] = -90.0f;
				} else if (M23 == 0.0) {
					if (M13 > 0.0)
						fchi[j] = 180.0f;
					else
						fchi[j] = 0.0f;
				} else {
					fchi[j] = - (float) (atan(M23 / M13) * PITODEG);
					if (M13 > 0.0)
						fchi[j] += 180.0f;
				}
		
				if (fomega[j] < 0.0) {
					fomega[j] = -fomega[j];
					fchi[j] += 180.0f;
				}
				if (fomega[j] > 90.0) {
					fomega[j] = 180.0f - fomega[j];
					fchi[j] += 180.0f;
				}
				while (fchi[j] > 360.0)
					fchi[j] -= 360.0f;
				while (fchi[j] < 0.0)
					fchi[j] += 360.0f;

			}
		}
		
      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalPut(env, omega, fomega);
		JNIArrayCriticalPut(env, chi, fchi);
		JNIArrayCriticalRel(env, phi, fphi);
		JNIArrayCriticalRel(env, eta, feta);
		JNIArrayCriticalRel(env, sampleAngles, fsampleAngles);
		
		free(vM);

		return 0;
	}
#endif
  return 1;
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    sComputeTextureAngles
 * Signature: ([F[F[F[F[FF)I
 */
																									
JNIEXPORT jint JNICALL Java_it_unitn_maud_JNIAltivec_sComputeTextureAngles
  (JNIEnv *env, jclass obj, jfloatArray omega, jfloatArray chi, jfloatArray phi, jfloatArray eta,
   jfloatArray sampleAngles, jfloat position) {
	
  jsize size;
  int i, j;
	
  float *fomega, *fchi, *fphi, *feta, *fsampleAngles;
	
	if(omega != NULL) {
		size = JNIGetArrayLength(env, omega);
				
		fomega = (float *)JNIArrayCriticalGet(env, omega);
		fchi = (float *)JNIArrayCriticalGet(env, chi);
		fphi = (float *)JNIArrayCriticalGet(env, phi);
		feta = (float *)JNIArrayCriticalGet(env, eta);
		fsampleAngles = (float *)JNIArrayCriticalGet(env, sampleAngles);
		
		position /= 2.0f;
		position *= DEGTOPI;
		for (i = 0; i < 3; i++)
			fsampleAngles[i] *= DEGTOPI;
		float fsinOmegaSample = (float) sin(fsampleAngles[0]);
		float fcosOmegaSample = (float) cos(fsampleAngles[0]);
		float fsinCsiSample = (float) sin(fsampleAngles[1]);
		float fcosCsiSample = (float) cos(fsampleAngles[1]);
		float fsinTheta = (float) sin(position);
		float fcosTheta = (float) cos(position);
		for (j = 0; j < size; j++) {

				fomega[j] += fsampleAngles[2];
				float sinOmega = sin(fomega[j]);
				float cosOmega = cos(fomega[j]);
				float sinPhi = sin(fphi[j]);
				float cosPhi = cos(fphi[j]);
				float sinCsi = sin(fchi[j]);
				float cosCsi = cos(fchi[j]);
				float sinEta = sin(feta[j]);
				float cosEta = cos(feta[j]);
		
				float c1 = fcosOmegaSample * cosPhi;
				float cs1 = fcosOmegaSample * sinPhi;
				float s2 = fsinOmegaSample * cosPhi;
				float c2 = fcosCsiSample * cosCsi;
				float c3 = fcosCsiSample * sinCsi;
				float c4 = fsinCsiSample * sinCsi;
				float s1 = fcosCsiSample * sinPhi;
				float const1 = c1 * fsinCsiSample + c2;
				float const2 = -const1 * sinCsi + fcosOmegaSample * c2;
				float const3 = cs1 * fsinCsiSample - s2;
				float const4 = const1 * cosCsi + fcosOmegaSample * c3;
				float const5 = s2 * fsinCsiSample - cs1;
				float const6 = -const5 * sinCsi + fsinOmegaSample * c2;
				float const7 = fsinOmegaSample * sinPhi * fsinCsiSample + c1;
				float const8 = const5 * cosCsi + fsinOmegaSample * c3;
				float const9 = -c3 * cosPhi - fsinCsiSample * cosCsi;
				float const10 = s1 * sinOmega + const9 * cosOmega;
				float const11 = c2 * cosPhi - c4;

				float M13 = -(s1 * cosOmega - const9 * sinOmega) * fsinTheta + 
										(const11 * sinEta + const10 * cosEta) * fcosTheta; // res[0][2];
				float M23 = -(const7 * cosOmega - const6 * sinOmega) * fsinTheta + (const8 * sinEta + 
										(const7 * sinOmega + const6 * cosOmega) * cosEta) * fcosTheta; // res[1][2];
				float cosPsi = -(const3 * cosOmega - const2 * sinOmega) * fsinTheta + (const4 * sinEta + 
											(const3 * sinOmega + const2 * cosOmega) * cosEta) * fcosTheta; // res[2][2];
		
				fomega[j] = (float) (acos(cosPsi) * PITODEG);
		
				if (fabs(M23) < 1.0E-9)
					M23 = 0.0f;
				if (fabs(M13) < 1.0E-9) {
					if (M23 >= 0.0)
						fchi[j] = 90.0f;
					else
						fchi[j] = -90.0f;
				} else if (M23 == 0.0) {
					if (M13 > 0.0)
						fchi[j] = 180.0f;
					else
						fchi[j] = 0.0f;
				} else {
					fchi[j] = - (float) (atan(M23 / M13) * PITODEG);
					if (M13 > 0.0)
						fchi[j] += 180.0f;
				}
		
				if (fomega[j] < 0.0) {
					fomega[j] = -fomega[j];
					fchi[j] += 180.0f;
				}
				if (fomega[j] > 90.0) {
					fomega[j] = 180.0f - fomega[j];
					fchi[j] += 180.0f;
				}
				while (fchi[j] > 360.0)
					fchi[j] -= 360.0f;
				while (fchi[j] < 0.0)
					fchi[j] += 360.0f;

		}
		
      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
		JNIArrayCriticalPut(env, omega, fomega);
		JNIArrayCriticalPut(env, chi, fchi);
		JNIArrayCriticalRel(env, phi, fphi);
		JNIArrayCriticalRel(env, eta, feta);
		JNIArrayCriticalRel(env, sampleAngles, fsampleAngles);

		return 0;
	}
  return 1;
  
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vdivfA
 * Signature: ([F[F[F)V
 */
JNIEXPORT void JNICALL Java_it_unitn_maud_JNIAltivec_vdivfA
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize size, bsize, vsize;
  int i;

	jsize shouldBe4 = sizeof(vector float) / sizeof(float);
	size = JNIGetArrayLength(env, fvA);
	vsize = size / shouldBe4;
	bsize = size * sizeof(float);
		
	vector float *vA, *vB, *vC, *vT;
	vector float *vM, *vMnext;
	float *fA, *fB, *fC;

	fA = (float *)JNIArrayCriticalGet(env, fvA);
	fB = (float *)JNIArrayCriticalGet(env, fvB);
	fC = (float *)JNIArrayCriticalGet(env, fvC);

        // Doing a single malloc and free avoids unnecessary overhead.
	vM = (vector float *)malloc(bsize + bsize + bsize);
	memcpy(vA = vM, fA, bsize);
	memcpy(vB = vA + vsize, fB, bsize);
	vC = vB + vsize;

        // Could unroll this loop since we likely have two vector units.
	for (i = 0; i < vsize; i++) {
		vC[i] = vdivf(vA[i], vB[i]);
	}

	memcpy(fC, vC, bsize);

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
	JNIArrayCriticalRel(env, fvA, fA);
	JNIArrayCriticalRel(env, fvB, fB);
	JNIArrayCriticalPut(env, fvC, fC);
	free(vM);
#endif

}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vdivfF
 * Signature: ([F[F[F)I
 */
JNIEXPORT jint JNICALL Java_it_unitn_maud_JNIAltivec_vdivfF
	(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  vector float *vA, *vB, *vC;
  jsize size, size4;
  int i;

  if (fvA != NULL && fvB != NULL && fvC != NULL) {
  	vA = (vector float *) JNIGetFloatArrayElements(env, fvA);
  	vB = (vector float *) JNIGetFloatArrayElements(env, fvB);
  	vC = (vector float *) JNIGetFloatArrayElements(env, fvC);
  	size = JNIGetArrayLength(env, fvA);

    size4 = size / 4;

    for (i = 0; i < size4; i++) {
      vC[i] = vdivf(vA[i], vB[i]);
    }

    JNIReleaseFloatArrayElements(env, fvA, (jfloat *)vA);
    JNIReleaseFloatArrayElements(env, fvB, (jfloat *)vB);
    JNISetFloatArrayRegion(env, fvC, size, (jfloat *)vC);
    JNIReleaseFloatArrayElements(env, fvC, (jfloat *)vC);

    return 0;
  }
#endif
  return 1;
}

/*
 * Class:     it_unitn_maud_JNIAltivec
 * Method:    vdivf2
 * Signature: ([F[F[F)I
 * Returns:   Zero to indicate success. Non-zero is an error code.
 */
JNIEXPORT jint JNICALL Java_it_unitn_maud_JNIAltivec_vdivf2
(JNIEnv *env, jclass obj, jfloatArray fvA, jfloatArray fvB, jfloatArray fvC) {
#ifdef MAC_OS_X
  jsize size, bsize, vsize, asize;
  int i;

  if (fvA != NULL && fvB != NULL && fvC != NULL) {
    size = JNIGetArrayLength(env, fvA);
    bsize = size * sizeof(float);
    vsize = bsize / sizeof(vector float);
    // Paranoid check. Better safe than sorry.
    if ((	size == vsize * sizeof(vector float) / sizeof(float)) && size == JNIGetArrayLength(env, fvB) && 
					size == JNIGetArrayLength(env, fvC)) {
      vector float *vA, *vB, *vC;
      vector float *vM, *vMnext;
      // Determine if the arrays are conveniently aligned already.
      float *fA = (float *)JNIArrayCriticalGet(env, fvA);
      float *fB = (float *)JNIArrayCriticalGet(env, fvB);
      float *fC = (float *)JNIArrayCriticalGet(env, fvC);

      // Assume they are for now ...
      vA = (vector float *)fA;
      vB = (vector float *)fB;
      vC = (vector float *)fC;

      // Figure out which ones we need to clone into a malloc'd block (guaranteed aligned).
      jsize cloneA = ((unsigned int)fA & 0xf == 0) ? 0 : bsize;
      jsize cloneB = ((unsigned int)fB & 0xf == 0) ? 0 : bsize;
      jsize cloneC = ((unsigned int)fC & 0xf == 0) ? 0 : bsize;
      
      asize = cloneA + cloneB + cloneC;
      if (asize == 0) {
        // Could unroll this loop since we likely have two vector units.
        for (i = 0; i < vsize; i++) {
          vC[i] = vdivf(vA[i], vB[i]);
        }        
      } else {
        // Doing a single malloc and free avoids unnecessary overhead.
        vMnext = vM = (vector float *)malloc(asize);
        if (cloneA > 0) {
          vA = vMnext;
          vMnext += vsize;
          memcpy(vA, fA, bsize);
        }
        if (cloneB > 0) {
          vB = vMnext;
          vMnext += vsize;
          memcpy(vB, fB, bsize);
        }
        if (cloneC > 0) {
          vC = vMnext;
        }

        // Could unroll this loop since we likely have two vector units.
        for (i = 0; i < vsize; i++) {
          vC[i] = vdivf(vA[i], vB[i]);
        }

        if (cloneC > 0) {
          memcpy(fC, vC, bsize);
        }
        free(vM);
      }

      // Return the computed data. Note that we only commit the data we changed.
      // This can improve performance on some Java VMs, though it makes no difference
      // to HotSpot when using ReleasePrimitiveArrayCritical.
      JNIArrayCriticalRel(env, fvA, vA);
      JNIArrayCriticalRel(env, fvB, vB);
      JNIArrayCriticalPut(env, fvC, vC);

      return 0;
    }
    // Should probably throw some form of exception
    return 2;
  }
#endif
  return 1;
}

